/*
 * Copyright (C) 2001 USAGI/WIDE Project.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $USAGI: spd.c,v 1.6 2001/08/24 05:05:28 miyazawa Exp $
 */

/*
 * spd.c provide manipulatoin routines for IPsec SPD.
 * struct ipsec_sp represent a policy in IPsec SPD.
 * struct ipsec_sp refers IPsec SA by struct sa_index.
 */

#ifdef __KERNEL__

#ifdef MODULE
#  include <linux/module.h>
#  ifdef MODVERSIONS
#    include <linux/modversions.h>
#  endif /* MODVERSIONS */
#endif /* MODULE */

#ifdef SPD_DEBUG
#define DEBUG 1
#endif

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/init.h>
#include <linux/net.h>
#include <linux/in.h>
#include <linux/in6.h>
#include <linux/malloc.h>
#include <linux/spinlock.h>
#include <linux/list.h>
#include <linux/socket.h> /* sa_family_t */
#include <linux/skbuff.h> /* sk_buff */

#include <net/spd.h>
#include <net/sadb.h>

#include "spd_utils.h"
#include "sadb_utils.h"

#ifdef CONFIG_PROC_FS
#include <linux/proc_fs.h>
#endif /* CONFIG_PROC_FS */

#endif /* __KERNEL */

#define BUFSIZE 64

/* 
 * sa_list represents SADB as a list.
 * sadb_lock is a lock for SADB.
 * If you access a element in SADB,
 * you must lock and unlock each element before access.
 *
 * 
 * struct list_head *pos = NULL;
 * struct ipsec_sp *tmp = NULL;
 *                                
 * read_lock(&spd_lock);
 * list_for_each(pos, &sadb_list){
 *      tmp = list_entry(pos, struct ipsec_sp, entry);
 *      read_lock(&(tmp->lock));
 *
 *      some operation
 *
 *      read_unlock(&(tmp->lock));
 * }
 * read_unlock(&spd_lock);
 *
 * ## If you want to change the list or a element,
 *    you read these code with write_lock, write_unlock
 *    instead of read_lock, read_unlock. 
 */



LIST_HEAD(spd_list);
rwlock_t spd_lock = RW_LOCK_UNLOCKED;

struct ipsec_sp* ipsec_sp_kmalloc()
{
	struct ipsec_sp* sp = NULL;

	sp = (struct ipsec_sp*)kmalloc(sizeof(struct ipsec_sp), GFP_KERNEL);

	if (!sp) {
		pr_debug("spd_kmallc: entry couldn\'t be allocated.\n");
		return NULL;
	}

	ipsec_sp_init(sp);

	return sp;
}

int ipsec_sp_init(struct ipsec_sp *policy)
{
	int error = 0;
	if (!policy) {
		pr_debug("ipsec_sp_init: policy is null\n");
		error = -EINVAL;
		goto err;
	}
	memset(policy, 0, sizeof(struct ipsec_sp));
	INIT_LIST_HEAD(&(policy->sa_index_list));
	policy->lock = RW_LOCK_UNLOCKED;
	atomic_set(&(policy->count),0);
err:
	return error;
}

int ipsec_sp_kfree(struct ipsec_sp **policy)
{
	int error = 0;

	if (!*policy) {
		pr_debug("ipsec_sp_kfree: entry is null\n");
		error = -EINVAL;
		goto err;
	}

	error = sa_list_clear(&((*policy)->sa_index_list));
	kfree(*policy);
err:
	return error;
}

int ipsec_sp_copy(struct ipsec_sp *dst, struct ipsec_sp *src)
{
	int error = 0;

	if (!(dst&&src)) {
		pr_debug("ipsec_sp_copy: dst or src is null\n");
		error = -EINVAL;
		goto err;
	}

	memcpy(&(dst->selector), &(src->selector), sizeof(struct selector));
	dst->action = src->action;
	sa_list_copy(&(dst->sa_index_list), &(src->sa_index_list));
err:
	return error;
}

int ipsec_sp_release_ref(struct ipsec_sp *policy)
{
	int error = 0;

	if (policy) {
		pr_debug("ipsec_sp_release_ref: policy is null\n");
		error = -EINVAL;
		goto err;
	}

	atomic_dec(&(policy->count));
	pr_debug("ipsec_sp_release_ref: ptr=%p, count=%d\n",
			policy, atomic_read(&(policy->count)));
err:
	return error;
}

int spd_append(struct ipsec_sp *policy)
{
	int error = 0;
	struct ipsec_sp *new = NULL;

	if (!policy) {
		pr_debug("spd_append: policy is null\n");
		error = -EINVAL;
		goto err;
	}

	new = ipsec_sp_kmalloc();
	if (!new) {
		pr_debug("spd_append: ipsec_sp_kmalloc failed\n");
		error = -ENOMEM;
		goto err;
	}

	error = ipsec_sp_init(new);
	if (error) {
		pr_debug("spd_append: ipsec_sp_init failed\n");
		goto err;
	}

	error = ipsec_sp_copy(new, policy);
	if (error) {
		pr_debug("spd_append: ipsec_sp_copy failed\n");
		goto err;
	}

	write_lock(&spd_lock);
	list_add_tail(&(new->entry), &spd_list);
	write_unlock(&spd_lock);
err:
	return error;
}

int spd_remove(struct selector *selector)
{
	int error = 0;
	struct ipsec_sp *policy = NULL;

	pr_debug("spd_remove:\n");
	if (!selector) {
		pr_debug("spd_remove: selector is null\n");
		error = -EINVAL;
		goto err;
	}

	error = spd_find_by_selector(selector, &policy);
	pr_debug("spd_remove: spd_find_by_selector done\n");
	if (error == -EEXIST) {
		write_lock(&spd_lock);
		list_del(&(policy->entry));
		write_unlock(&spd_lock);
		error = ipsec_sp_kfree(&policy);
	}

err:
	pr_debug("spd_remove: error=%u\n", error);
	return error;
}

int spd_update(struct ipsec_sp *policy)
{
	int error = 0;
	struct ipsec_sp **tmp = NULL;

	if (!policy) {
		pr_debug("spd_update: policy is null\n");
		error = -EINVAL;
		goto err;
	}

	error = spd_find_by_selector(&(policy->selector), tmp);

	if (error != -EEXIST) {
		pr_debug("spd_update: spd_find_by_selector faild\n");
		goto err;
	}

	memcpy(&((*tmp)->selector), &(policy->selector), sizeof(struct selector));
	(*tmp)->action = policy->action;
	sa_list_copy(&((*tmp)->sa_index_list), &(policy->sa_index_list));

err:
	return error;
}

int spd_find_by_selector(struct selector *selector, struct ipsec_sp **policy)
{
	int error = -ESRCH;
	struct list_head *pos = NULL;
	struct ipsec_sp *tmp = NULL;

	if (!selector) {
		pr_debug("spd_find_by_selector: selector is null\n");
		error = -EINVAL;
		goto err;
	}

	read_lock(&spd_lock);
	list_for_each(pos, &spd_list){
		tmp = list_entry(pos, struct ipsec_sp, entry);
		read_lock(&(tmp->lock));
		if (!compare_selector(selector, &(tmp->selector))) {
			pr_debug("spd_find_by_selector: found matched element\n");
			error = -EEXIST;
			*policy = tmp;
			pr_debug("spd_find_by_selector: copy tmp to *policy \n");
			read_unlock(&(tmp->lock));
			break;
		}
		read_unlock(&(tmp->lock));
	}
	read_unlock(&spd_lock);

err:
	return error;
}

int spd_find_by_skbuff(struct sk_buff *skb, sa_family_t sa_family, struct ipsec_sp *policy)
{
	int error = -ESRCH;
	struct list_head *pos;
	struct ipsec_sp *tmp;

	if (!skb) {
		pr_debug("spd_find_by_skbuff: skb is null\n");
		error = -EINVAL;
		goto err;
	}

	read_lock(&spd_lock);
	list_for_each(pos, &spd_list){
		tmp = list_entry(pos, struct ipsec_sp, entry);
		read_lock(&(tmp->lock));
		if (!compare_skb_selector(skb, sa_family, &(tmp->selector))) {
			error = -EEXIST;
			policy = tmp;
		}
		read_unlock(&(tmp->lock));
	}
	read_unlock(&spd_lock);

err:
	return error;
}
		
int spd_check_sa_list(struct ipsec_sp *policy)
{
	int error = 0;
	struct list_head *pos;
	struct sa_index *sa_index;

	if (!policy) {
		pr_debug("spd_check_sa_list: policy is null\n");
		error = -EINVAL;
		goto err;
	}

	list_for_each(pos, &(policy->entry)){
		sa_index = list_entry(pos, struct sa_index, entry);
		if (sa_index->sa == NULL) {
			error = sadb_find_by_sa_index(sa_index);
			if (error == -EINVAL) {

			} else if (error == -ESRCH) {
				error = -ENODATA;
				goto err;
			/* TODO:
				If sa_index has NULL pointer as sa,
				it call SADB_AQCURE message to KMd in user land.
			*/
			} else {
				goto err;
			}
		}
	}

err:
	return error;
}

void spd_clear_db()
{
	struct list_head *pos;
	struct ipsec_sp *policy;

	write_lock(&spd_lock);
	list_for_each(pos, &spd_list){
		policy = list_entry(pos, struct ipsec_sp, entry);
		list_del(&(policy->entry));
		ipsec_sp_kfree(&policy);		
	}
	write_unlock(&spd_lock);
}


#ifdef CONFIG_PROC_FS
static int spd_get_info(char *buffer, char **start, off_t offset, int length)
{
	int error = 0;
	int count = 0;
	int len = 0;
        off_t pos=0;
        off_t begin=0;
        char buf[BUFSIZE]; 
        struct list_head *list_pos = NULL;
	struct list_head *sa_list_pos = NULL;
        struct ipsec_sp *tmp_spd = NULL;
	struct sa_index *tmp_sa = NULL;
	
        len += sprintf(buffer + len, "---------------------------------------------\n"); 
        read_lock(&spd_lock);
        list_for_each(list_pos, &spd_list){
#if 0
		pr_debug("spd_get_info: list_pos->prev=%p\n", list_pos->prev);
		pr_debug("spd_get_info: list_pos->next=%p\n", list_pos->next);
#endif
		count = 0;
                tmp_spd = list_entry(list_pos, struct ipsec_sp, entry);
                memset(buf, 0, BUFSIZE);
                sockaddrtoa((struct sockaddr*)&(tmp_spd->selector.src), buf, BUFSIZE);
                len += sprintf(buffer + len, "src:%s\n", buf);
                memset(buf, 0, BUFSIZE);
                sockaddrtoa((struct sockaddr*)&(tmp_spd->selector.dst), buf, BUFSIZE);
                len += sprintf(buffer + len, "dst:%s\n", buf);
		len += sprintf(buffer + len, "proto:%u\n", tmp_spd->selector.proto);
                len += sprintf(buffer + len, "action:%u\n", tmp_spd->action);

		list_for_each(sa_list_pos, &(tmp_spd->sa_index_list)){
			tmp_sa = list_entry(sa_list_pos, struct sa_index, entry);
			len += sprintf(buffer + len, "sa[%d]: ", count);
			sockaddrtoa((struct sockaddr*)&(tmp_sa->dst), buf, BUFSIZE);
			len += sprintf(buffer + len, "sa->dst:%s\n", buf);
			len += sprintf(buffer + len, "sa[%d]: sa->ipsec_proto:%u ", count++, htons(ntohl(tmp_sa->ipsec_proto)));
			len += sprintf(buffer + len, "sa->spi:0x%x\n", htonl(tmp_sa->spi));
		}
                len += sprintf(buffer + len, "---------------------------------------------\n"); 

                pos=begin+len;
                if (pos<offset) {
                        len=0;
                        begin=pos;
                }
                if (pos>offset+length) {
                        read_unlock(&spd_lock);
                        goto done;
                }
        }       
        read_unlock(&spd_lock);
done:

        *start=buffer+(offset-begin);
        len-=(offset-begin);
        if (len>length)
                len=length;
        if (len<0)
                len=0;
        return len;

	goto err;
err:
	return error;
}
#endif /* CONFIG_PROC_FS */

int spd_init(void)
{
        int error = 0;

	INIT_LIST_HEAD(&spd_list);
	pr_debug("spd_init: spd_list.prev=%p\n", spd_list.prev);
	pr_debug("spd_init: spd_list.next=%p\n", spd_list.next);
#ifdef CONFIG_PROC_FS
        proc_net_create("spd", 0, spd_get_info);
#endif /* CONFIG_PROC_FS */

	pr_info("spd_init: SPD initialized\n");
        return error;
}

int spd_cleanup(void)
{
        int error = 0;

#ifdef CONFIG_PROC_FS
        proc_net_remove("spd");
#endif /* CONFIG_PROC_FS */

        spd_clear_db();

	pr_info("spd_cleanup: SPD cleaned up\n");
        return error;
}

#if 0

/* #ifdef MODULE */

static int __init spd_module_init(void)
{
        int err = spd_init();
        return err;
}

static void __exit spd_module_cleanup(void)
{
        spd_cleanup();
}

module_init(spd_module_init);
module_exit(spd_module_cleanup);


EXPORT_SYMBOL(ipsec_sp_kmalloc);
EXPORT_SYMBOL(ipsec_sp_init);
EXPORT_SYMBOL(ipsec_sp_kfree);
EXPORT_SYMBOL(spd_append);
EXPORT_SYMBOL(spd_remove);
EXPORT_SYMBOL(spd_update);
EXPORT_SYMBOL(spd_clear_db);
EXPORT_SYMBOL(spd_find_by_selector);


#endif /* MODULE */

